using LinearAlgebra
using SparseArrays
using FileIO
using ImageIO
using ImageShow
using Images
using Colors

function ForwardDifferences1D(dim, mirror)
    """
    Forward finite differences matrix (sparse) of a 1D signal of length dim.
    If mirror is true, then the first and last entry of the signal are assumed
    to extend infinitely on their respective side.
    """
    if mirror
        last_entry = dim - 1
    else
        last_entry = dim
    end
    ir = vcat(collect(1:last_entry), collect(1:(dim - 1)))
    jc = vcat(collect(1:last_entry), collect(2:dim))
    va = vcat(-1.0 .* ones(last_entry), ones(dim - 1))
    return sparse(ir, jc, va, dim, dim)
end

function BackwardDifferences1D(dim, mirror)
    """
    Backward finite differences matrix (sparse) of a 1D signal of length dim.
    If mirror is true, then the first and last entry of the signal are assumed
    to extend infinitely on their respective side.
    """
    if mirror
        first_entry = 2
    else
        first_entry = 1
    end
    ir = vcat(collect(first_entry:dim), collect(2:dim))
    jc = vcat(collect(first_entry:dim), collect(1:(dim - 1)))
    va = vcat(ones(dim + 1 - first_entry), -1.0 .* ones(dim - 1))
    return sparse(ir, jc, va, dim, dim)
end

function SecondDifferences1D(dim, mirror)
    """
    Second order finite differences matrix (sparse) of a 1D signal of length
    dim. If mirror is true, then the first and last entry of the signal are
    assumed to extend infinitely on their respective side.
    """
    return ForwardDifferences1D(dim, mirror) -
        BackwardDifferences1D(dim, mirror)
end

function Laplacian2D(xdim, ydim, mirror)
    """
    Standard finite difference matrix (sparse) for the Laplacian in 2D. The
    signal has size xdim in x-direction and ydim in y-direction. If mirror is
    true, then the first and last entry along each dimension are assumed to
    extend infinitely on their respective side.
    """
    return kron(SecondDifferences1D(ydim, mirror), sparse(I, xdim, xdim)) +
        kron(sparse(I, ydim, ydim), SecondDifferences1D(xdim, mirror))
end

function InpaintingMatrix(xdim, ydim, C, mirror)
    """
    Build the sparse system matrix for PDE-based inpainting with the Laplacian
    using a standard finite difference scheme. The image should have xdim
    columns and ydim rows. If mirror is true, then Neumann boundary conditions
    are used, otherwise it will be 0 Dirichlet boundary conditions.
    C should be a (sparse) diagonal matrix. Each entry on the diagonal should
    contain the weight for the inpainting of the respective pixel. The linear
    indexing of the pixels is assumed to be columnwise.
    """
    return C + (sparse(I, xdim * ydim, xdim * ydim) - C) *
        (-1.0 * Laplacian2D(xdim, ydim, mirror))
end

# Read in image and mask data.
input_image = load("lena.png")
mask_image = load("lena-mask-5.png")
# Set up mask C, the system matrix and the righthand side.
nr, nc = size(input_image)
mask_pos = findall(!iszero, vec(Int8.(mask_image)))
nz, = size(mask_pos)
C = sparse(mask_pos, mask_pos, vec(Float64.(mask_image))[mask_pos], nr * nc, nr * nc)
SystemMatrix = InpaintingMatrix(nc, nr, C, true)
RighthandSide = vec(Float64.(input_image .* mask_image))
# Solve the linear system using the backslash operator
Reconstruction = SystemMatrix \ RighthandSide
# Reshape result into an image.
InpaintingResult = reshape(Reconstruction, (nr, nc))
colorview(Gray, InpaintingResult)